//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// October 2000
//
#include <Epproc.h>

#include <Dal.h>	// SAS
#include <OalServer.h>
#include <CalServer.h>
#include <param.h>
#include <errstr.h>
#include <metatask.h>
#include <Proc.h>
#include <Evaluator.h>

#include <attcalc.h>     // Tasks
#include <atthkgen.h>
#include <badpix.h>
#include <badpixfind.h>
#include <epframes.h>
#include <epevents.h>
#include <evlistcomb.h>
#include <hkgtigen.h>
#include <epauxcomb.h>
#include <timeappend.h>
#include <tabgtigen.h>
#include <gtimerge.h>
#include <gtialign.h>
#include <evselect.h>
#include <dscopyblock.h>
#include <dsrm.h>
#include <dsrename.h>

#include <algorithm>
#include <cstdlib>
#include <unistd.h>

Epproc::Epproc() :
  _mypars(*Info::Task::current()->parameters()),
  _inst(xmmPayLoad.instrument(Xmm::EPN)),
  _hasfilterexpression(false),
  _hasatthkgenset(false),
  _hasobservationlevelgti(false),
  _hashkgti(false),
  _proc(Proc("./"))
{
  PROC

  oalServer.instrument(_inst.id());
 
  if(_mypars["removebadevents"]->boolValue()){
    _filterexpression = "#XMMEA_EP";
    _hasfilterexpression = true;
  }
  if(_mypars["withfilterexpression"]->boolValue()){
    _filterexpression += " && " +
      _mypars["filterexpression"]->stringValue();
    _hasfilterexpression = true;
  }

  if(_mypars["withgtiset"]->boolValue())
    _gtis.push_back(_mypars["gtiset"]->stringValue());

  ENDPROC
}  

void Epproc::selectDataModes()
{
  PROC
  if(_mypars["selectmodes"]->boolValue()){
    if(_mypars["imaging"]->boolValue())
      _datamodes.push_back(Xmm::IMAGING);
    if(_mypars["burst"]->boolValue())
      _datamodes.push_back(Xmm::BURST);
    if(_mypars["timing"]->boolValue())
      _datamodes.push_back(Xmm::TIMING);
  } else 
    _datamodes.push_back(Xmm::IMAGING);

  for(unsigned k=0; k<_datamodes.size(); ++k)
    errstr << "Selected data mode: " << _datamodes[k]
	   << Sas::message(Msg::Meta, Msg::Silent);
  ENDPROC
}    


void Epproc::selectCcds()
{
  PROC
  if(_mypars["selectccds"]->boolValue()){
    if(_mypars["ccd1"]->boolValue())
      _ccds.push_back(1);
    if(_mypars["ccd2"]->boolValue())
      _ccds.push_back(2);
    if(_mypars["ccd3"]->boolValue())
      _ccds.push_back(3);
    if(_mypars["ccd4"]->boolValue())
      _ccds.push_back(4);
    if(_mypars["ccd5"]->boolValue())
      _ccds.push_back(5);
    if(_mypars["ccd6"]->boolValue())
      _ccds.push_back(6);
    if(_mypars["ccd7"]->boolValue())
      _ccds.push_back(7);
    if(_mypars["ccd8"]->boolValue())
      _ccds.push_back(8);
    if(_mypars["ccd9"]->boolValue())
      _ccds.push_back(9);
    if(_mypars["ccd10"]->boolValue())
      _ccds.push_back(10);
    if(_mypars["ccd11"]->boolValue())
      _ccds.push_back(11);
    if(_mypars["ccd12"]->boolValue())
      _ccds.push_back(12);
  } else
    for(unsigned k=1; k<=12; ++k)
      _ccds.push_back(k);

  for(unsigned k=0; k<_ccds.size(); ++k)
    errstr << "Selected CCD: " << _ccds[k]
	   << Sas::message(Msg::Meta, Msg::Silent);
  ENDPROC
}

void Epproc::selectExposures()
{
  PROC
  if(_mypars["selectexposures"]->boolValue())
    errstr << "exposure selection not implemented. All exposures will be processed." << Sas::warning("NotImplemented");

  for(unsigned k=0; k<_datamodes.size(); ++k){
    oalServer.dataMode(_datamodes[k]);
    unsigned n = oalServer.odf()->exposures();
    for(unsigned j=1; j<=n ; ++j){
      oalServer.exposureIndex(j);
      Proc::ScienceExposure s;
      s.mode = _datamodes[k];
      s.index = j;
      if(oalServer.odf()->hasSet(Xmm::AUXILIARY)){
	s.auxset.set = oalServer.odf()->selectSet(Xmm::AUXILIARY);
	s.hasauxset = true;
      } else
	s.hasauxset = false;
      s.exposure = oalServer.odf()->currentExposure();
      _exposures.push_back(s);
    }
  }

  for(unsigned k=0; k<_exposures.size(); ++k)
    errstr << "Selected exposure: " << _exposures[k].exposure->counter()
	   << Sas::message(Msg::Meta, Msg::Silent);
  ENDPROC
}


unsigned Epproc::numberOfExposures() const
{
  PROC
  RETURN _exposures.size();
}

void Epproc::processAttitude()
{
  PROC
  string rp = _mypars["referencepointing"]->stringValue();
  if(_mypars["runatthkgen"]->boolValue() || rp == "mean"
     || rp == "median"){
    Info::Task* atthkgen = task("atthkgen");
    ParameterInfo& atthkgenpars = *atthkgen->parameters();
    atthkgenpars.revertToDefault();
    atthkgenpars["atthkset"]->setValue(_proc.atthkgenname());
    if(_proc.protectedrun(atthkgen))
      _hasatthkgenset = true;
  }

 ENDPROC
}

void Epproc::processHk()
{
  PROC
  if(_mypars["runhkgtigen"]->boolValue()){
    Info::Task* hkgtigen = task("hkgtigen");
    ParameterInfo& hkgtigenpars = *hkgtigen->parameters();
    hkgtigenpars.revertToDefault();
    hkgtigenpars["instrument"]->setValue("epn");
    hkgtigenpars["withparameters"]->setValue(_mypars["withparameters"]);
    hkgtigenpars["parameters"]->setValue(_mypars["parameters"]);
    hkgtigenpars["except"]->setValue(_mypars["except"]);
    hkgtigenpars["withoverrideparameters"]->setValue(_mypars["withoverrideparameters"]);
    hkgtigenpars["overrideparameters"]->setValue(_mypars["overrideparameters"]);
    hkgtigenpars["withgtiset"]->setValue(true);
    hkgtigenpars["gtiset"]->setValue(_proc.hkgtiname());
    if(_proc.protectedrun(hkgtigen)){
      _gtis.push_back(_proc.hkgtiname());
      _hashkgti = true;
    }
  }
  
  ENDPROC
}

void Epproc::mergeGtis()
{
  PROC
  if(_gtis.size() > 0){
    Info::Task* gtimerge = task("gtimerge");
    ParameterInfo& gtimergepars = *gtimerge->parameters();
    gtimergepars.revertToDefault();
    gtimergepars["gtitable"]->setValue(_proc.observationgtiname() + ":STDGTI");
    gtimergepars["mergemode"]->setValue("or");
    vector<string> tables(_gtis.size());
    for(unsigned j=0; j<_gtis.size(); ++j)
      tables[j] = _gtis[j] + ":";
    gtimergepars["tables"]->setValue(tables);
    if(_proc.protectedrun(gtimerge))
      _hasobservationlevelgti = true;
  }
  
  ENDPROC
}


void Epproc::processExposure(unsigned& n)
{
  PROC
  Proc::ScienceExposure& thisexp = _exposures[n];

  vector<string> egtis; // exposure level GTIs
  vector<string> eventlists; // event lists to be merged
  vector<string> filteredeventlists; // filtered event lists to be merged

  errstr << "Processing ";
  _proc.print(thisexp, errstr);
  errstr << Sas::message(Msg::Meta, Msg::Silent);

  // --------------
  // Pre processing
  // --------------

  // AUX set processing
  // ------------------
  if(_mypars["withauxgtiexpression"]->boolValue() &&
     thisexp.hasauxset){
    
    errstr << "Processing the auxiliary file: "
	   << thisexp.auxset.set->name()
	   << Sas::message(Msg::Meta, Msg::Silent);
    
    Info::Task* epauxcomb = task("epauxcomb");
    ParameterInfo& epauxcombpars = *epauxcomb->parameters();
    epauxcombpars.revertToDefault();
    epauxcombpars["set"]->setValue(thisexp.auxset.set->name());
    epauxcombpars["outset"]->setValue(_proc.epauxcombname());
    if(_proc.protectedrun(epauxcomb)){
      Info::Task* timeappend = task("timeappend");
      ParameterInfo& timeappendpars = *timeappend->parameters();
      timeappendpars.revertToDefault();
      timeappendpars["table"]->setValue(_proc.epauxcombname()+":");
      timeappendpars["withsettings"]->setValue(true);
      timeappendpars["datamode"]->setValue(Xmm::toString(thisexp.mode));
      timeappendpars["expnr"]->setValue(thisexp.index);
      timeappendpars["instrument"]->setValue(Xmm::toString(_inst.id()));
      timeappendpars["outset"]->setValue(_proc.timeappendname());
      if(_proc.protectedrun(timeappend)){
	Info::Task* tabgtigen = task("tabgtigen");
	ParameterInfo& tabgtigenpars = *tabgtigen->parameters();
	tabgtigenpars.revertToDefault();
	tabgtigenpars["table"]->setValue(_proc.timeappendname()+":");
	tabgtigenpars["expression"]->setValue(_mypars["auxgtiexpression"]);
	tabgtigenpars["gtiset"]->setValue(_proc.auxgtiname());
	if(_proc.protectedrun(tabgtigen))
	  egtis.push_back(_proc.auxgtiname());
      }
    }
  }  

  const list<const OdfSet*>& l = thisexp.exposure->sets();
  
  // FIXME: this string contains the global selection expression that
  // must be applied to the merged event list so that the DSS
  // information is preserved. Ideally there would be a way to do this
  // in some other manner (via the DSS itself or evlistcomb)
  // In any case this should not be a member, as it has nothing to do
  // with the state of Emproc.
  _post_selection_expression = "";

  // ---------------
  // Main processing
  // ---------------

  // for each set in the exposure
  // ----------------------------  
  for(LI i=l.begin(); i!=l.end(); ++i){
    const OdfSet* os = (*i);
    if(find(_ccds.begin(), _ccds.end(), os->ccdNr()) != _ccds.end()){

      oalServer << os->name();

      vector<string> cgtis; // CCD level GTIs

      errstr << "Processing data set: " << os->name()
	 << Sas::message(Msg::Meta, Msg::Silent);
      
      Info::Task* epframes = task("epframes");
      ParameterInfo& epframespars = *epframes->parameters();    
      epframespars.revertToDefault();
      epframespars["gtiset"]->setValue(_proc.framesgtiname());
      epframespars["eventset"]->setValue(_proc.epframesname());

      // Parameters transferred from the meta-task
      epframespars["mipthreshold"]->setValue(_mypars["mipthreshold"]);
      epframespars["mipmethod"]->setValue(_mypars["mipmethod"]);
      epframespars["mipdiscard"]->setValue(_mypars["mipdiscard"]);
      epframespars["mipthreshold"]->setValue(_mypars["mipthreshold"]);
      epframespars["mipdist"]->setValue(_mypars["mipdist"]);
      epframespars["wrongpixlimit"]->setValue(_mypars["wrongpixlimit"]);
      
      epframespars["set"]->setValue(os->name());
      if(os->dataMode() == Xmm::TIMING ||
	 os->dataMode() ==Xmm::BURST){
	if(_mypars["withsrccoords"]->boolValue()){
	  epframespars["srcra"]->setValue(_mypars["srcra"]);
	  epframespars["srcdec"]->setValue(_mypars["srcdec"]);
	}
	if(os->dataMode() == Xmm::TIMING)
	  epframespars["srcposition"]->setValue(_mypars["timingsrcposition"]);
	else if(os->dataMode() == Xmm::BURST)
	  epframespars["srcposition"]->setValue(_mypars["burstsrcposition"]);
      }

      bool hasbadpixset = false;
      if(!_proc.protectedrun(epframes, true)){
	Sas::message(Msg::Meta, Msg::Silent,
		     os->name() + string(" Will not be further processed."));
	continue;
      }

      cgtis.push_back(_proc.framesgtiname());
    
      if(_mypars["runbadpixfind"]->boolValue()){
	Info::Task* badpixfind = task("badpixfind");
	ParameterInfo& badpixfindpars = *badpixfind->parameters();
	badpixfindpars.revertToDefault();
	badpixfindpars["eventset"]->setValue(_proc.epframesname());
	badpixfindpars["searchbadpix"]->setValue(_mypars["searchbadpixels"]);
	badpixfindpars["thresholdlabel"]->setValue(_mypars["thresholdlabel"]);
	badpixfindpars["lothresh"]->setValue(_mypars["lothresh"]);
	badpixfindpars["hithresh"]->setValue(_mypars["hithresh"]);
	badpixfindpars["columnsearchlabel"]->setValue(_mypars["columnsearchlabel"]);
	badpixfindpars["locolthresh"]->setValue(_mypars["locolthresh"]);
	badpixfindpars["hicolthresh"]->setValue(_mypars["hicolthresh"]);
	badpixfindpars["flickertimesteps"]->setValue(_mypars["flickertimesteps"]);
	badpixfindpars["flickerksthresh"]->setValue(_mypars["flickerksthresh"]);
	badpixfindpars["flickerchisqthresh"]->setValue(_mypars["flickerchisqthresh"]);
	badpixfindpars["cellsize"]->setValue(_mypars["cellsize"]);
	badpixfindpars["backgroundrate"]->setValue(_mypars["backgroundrate"]);
	badpixfindpars["narrowerthanpsf"]->setValue(_mypars["narrowerthanpsf"]);
	badpixfindpars["rawxlist"]->setValue(_mypars["rawxlist"]);
	badpixfindpars["rawylist"]->setValue(_mypars["rawylist"]);
	badpixfindpars["typelist"]->setValue(_mypars["typelist"]);
	badpixfindpars["yextentlist"]->setValue(_mypars["yextentlist"]);
	badpixfindpars["badpixset"]->setValue(_proc.badpixfindname());
	if(!_proc.protectedrun(badpixfind)){
	  Sas::message(Msg::Meta, Msg::Silent,
		       os->name() + string(" Will not be further processed."));
	  continue;
	}
	hasbadpixset = true;
      }

      Info::Task* badpix = task("badpix");
      ParameterInfo& badpixpars = *badpix->parameters();
      badpixpars.revertToDefault();
      badpixpars["badpixset"]->setValue(_proc.badpixfindname());
      badpixpars["eventset"]->setValue(_proc.epframesname());
      badpixpars["getuplnkbadpix"]->setValue(_mypars["getuplnkbadpix"]);
      badpixpars["getotherbadpix"]->setValue(_mypars["getotherbadpix"]);
      badpixpars["getnewbadpix"]->setValue(_mypars["getnewbadpix"]->boolValue()
					   && hasbadpixset);
      badpixpars["windowfilter"]->setValue(_mypars["windowfilter"]);
      if(!_proc.protectedrun(badpix)){
	Sas::message(Msg::Meta, Msg::Silent,
		     os->name() + string(" Will not be further processed."));
	continue;
      }


      Info::Task* epevents = task("epevents");
      ParameterInfo& epeventspars = *epevents->parameters();
      epeventspars.revertToDefault();
      epeventspars["randomizeposition"]->setValue(_mypars["randomizeposition"]);
      epeventspars["randomizeenergy"]->setValue(_mypars["randomizeenergy"]);
      epeventspars["gainctiaccuracy"]->setValue(_mypars["gainctiaccuracy"]);
      epeventspars["reemissionthresh"]->setValue(_mypars["reemissionthresh"]);
      epeventspars["withoutoftime"]->setValue(_mypars["withoutoftime"]);
      epeventspars["eventset"]->setValue(_proc.epframesname());
      epeventspars["outset"]->setValue(_proc.eventsname());
      if(!_proc.protectedrun(epevents, true)){
	Sas::message(Msg::Meta, Msg::Silent,
		     os->name() + string(" Will not be further processed."));
	continue;
      }
      
      Info::Task* attcalc = task("attcalc");
      ParameterInfo& attcalcpars = *attcalc->parameters();
      attcalcpars.revertToDefault();
      attcalcpars["withatthkset"]->setValue(_mypars["runhkgtigen"]);
      attcalcpars["atthkset"]->setValue(_proc.atthkgenname());
      attcalcpars["eventset"]->setValue(_proc.eventsname());
      string rp = _mypars["referencepointing"]->stringValue();
      if(rp == "nominal")
	attcalcpars["refpointlabel"]->setValue("nom");
      else if(rp == "object")
	attcalcpars["refpointlabel"]->setValue("obj");
      else if(rp == "mean"){
	attcalcpars["refpointlabel"]->setValue("pnt");
	attcalcpars["withmedianpnt"]->setValue(false);
      }
      else if(rp == "median"){
	attcalcpars["refpointlabel"]->setValue("pnt");
	attcalcpars["withmedianpnt"]->setValue(true);
      }
      else if(rp == "user"){
	attcalcpars["fixeddec"]->setValue(_mypars["dec"]);
	attcalcpars["fixedra"]->setValue(_mypars["ra"]);
	attcalcpars["fixedposangle"]->setValue(_mypars["posangle"]);
      }
      if(!_proc.protectedrun(attcalc)){
	Sas::message(Msg::Meta, Msg::Silent,
		     os->name() + string(" Will not be further processed."));
	continue;
      }
	
      // Add the observation-level and exposure-level GTIs to the
      // list of CCD-level GTIs.
      bool enablegtiprocessing = false;
      for(unsigned j=0; j<egtis.size(); ++j)
	cgtis.push_back(egtis[j]);      
      for(unsigned j=0; j<_gtis.size(); ++j)
	cgtis.push_back(_gtis[j]);
      if(cgtis.size() > 0){
	Info::Task* gtimerge = task("gtimerge");
	ParameterInfo& gtimergepars = *gtimerge->parameters();
	gtimergepars.revertToDefault();
	gtimergepars["withgtitable"]->setValue(true);
	gtimergepars["gtitable"]->setValue(_proc.ccdgtitablename(true));
	//	  gtimergepars["gtitable"]->setValue(_proc.eventsname() + ":STDGTI");
	gtimergepars["mergemode"]->setValue("and");
	vector<string> tables(cgtis.size());
	for(unsigned j=0; j<cgtis.size(); ++j)
	  tables[j] = cgtis[j] + ":";
	if(_hasobservationlevelgti)
	  tables.push_back(_proc.observationgtiname() + ":STDGTI");
	gtimergepars["tables"]->setValue(tables);
	if(_proc.protectedrun(gtimerge)){
	  enablegtiprocessing = true;
	}
      }


      // Also rename the EXPOSURE extension to EXPOSUnn, so the
      // eveselect can update the exposure information, and the ccd
      // level event list can be used as is.
      //
      Info::Task* dsrename = task("dsrename");
      ParameterInfo& dsrenamepars = *dsrename->parameters();
      dsrenamepars.revertToDefault();
      dsrenamepars["object"]->setValue(_proc.eventsname() +
				       ":EXPOSURE");

      // FIXME: sstr is also used in the following if-block.
      strstream sstr;
      sstr << "EXPOSU" << setw(2) << setfill('0')
	   << oalServer.ccd() << ends;
      dsrenamepars["newname"]->setValue(sstr.str());
      if(!_proc.protectedrun(dsrename)){
	Sas::message(Msg::Meta, Msg::Silent,
		     os->name() + string(" Will not be further processed."));
	continue;
      }

      // Remove a number of columns from the exposure extension that
      // are not used.
      if(_mypars["deleteexposurecolumns"]->boolValue()){
	string s(_proc.eventsname() + ":" + sstr.str());
	vector<string> ss(5,s);
	ss[0] += ":FRAMELIM";
	ss[1] += ":NDSCLIN1";
	ss[2] += ":NDSCLIN2";
	ss[3] += ":NDSCLIN3";
	ss[4] += ":NDSCLIN4";
	Info::Task *dsrm = task("dsrm");
	ParameterInfo& dsrmpars = *dsrm->parameters();
	dsrmpars.revertToDefault();
	dsrmpars["objects"]->setValue(ss);
	if(!_proc.protectedrun(dsrm))
	  Sas::message(Msg::Meta, Msg::Sparse, "Could not remove extra columns from " + s);       
      }
      
      if(_mypars["applygti"]->boolValue() && enablegtiprocessing){
	strstream s;
	s << " CCDNR == " << oalServer.ccd() << " && TIME in gti("
	  << _proc.ccdgtitablename(true) << ")" << ends;
	// FIXME: I would like to filter on GTI here as well but I
	// cannot.
	//	expression += string(" && ") + s.str();
	if(_post_selection_expression.size() > 0)
	  _post_selection_expression += "||";
	else
	  _post_selection_expression = "(";
	_post_selection_expression += string("(") + s.str() + string(")");
      }
      
      // Here we only filter with an event removing filter. GTI can be
      // done only on the combined event list.
      if(_hasfilterexpression){
	string expression = _filterexpression;
	Info::Task* evselect = task("evselect");
	ParameterInfo& evselectpars = *evselect->parameters();
	evselectpars.revertToDefault();
	evselectpars["table"]->setValue(_proc.eventsname() + ":EVENTS");
	evselectpars["withfilteredset"]->setValue(true);
	evselectpars["filteredset"]->setValue(_proc.eventsname(true));
	evselectpars["destruct"]->setValue(true);
	evselectpars["expression"]->setValue(expression);
	evselectpars["writedss"]->setValue(true);
	//	evselectpars["updateexposure"]->setValue(true);
	evselectpars["updateexposure"]->setValue(false);
	evselectpars["keepfilteroutput"]->setValue(true);
	if(!_proc.protectedrun(evselect)){
	  Sas::message(Msg::Meta, Msg::Silent,
		       os->name() + string(" Will not be further processed."));
	  continue;
	}
	filteredeventlists.push_back(_proc.eventsname(true));
      }
      eventlists.push_back(_proc.eventsname());      
    } else // CCD loop
      errstr << os->name()
	     << " not processed as it does not correspond to a selected CCD."
	     << Sas::message(Msg::Meta, Msg::Sparse);
  } // for each set in the exposure

  if(_post_selection_expression.size() > 0)
    _post_selection_expression += ")";

  // FIXME (code duplication)
  // combineEventLists will remove the intermediate event lists if
  // necessary. Delete here as well in case createunfilteredeventlists
  // is false.
  if(_mypars["createunfilteredeventlists"]->boolValue())
    combineEventLists(eventlists, thisexp, false);
  else if(_mypars["removeintermediateeventlists"]->boolValue())
    for(unsigned k = 0; k < eventlists.size(); ++k)
      unlink(eventlists[k].c_str());
  
  
  combineEventLists(filteredeventlists, thisexp, true);

  // Before going on to the next exposure do whatever clean up is required
  cleanup();
  
  ENDPROC
} 

void Epproc::combineEventLists(const vector<string>& el,
			       const Proc::ScienceExposure& thisexp,
			       const bool& filtered = false)
{
  PROC
  string eventlist;
  bool removeeventlists = _mypars["removeintermediateeventlists"]->boolValue();
  if(_mypars["runevlistcomb"]->boolValue()){
    if(el.size() == 0){
      _proc.print(thisexp, errstr);
      errstr << Sas::warning("NoEventListsToMerge");
    } else {
      Info::Task* evlistcomb = task("evlistcomb");
      ParameterInfo& evlistcombpars = *evlistcomb->parameters();      
      evlistcombpars.revertToDefault();
      evlistcombpars["instrument"]->setValue("epn");
      evlistcombpars["eventsets"]->setValue(el);
      
      // FIXME: the user of evlistcomb is _not_ supposed to change the
      // value of the parameter "emosdatamodes". See SSC-SPR-2392.
//        vector<string> datamodes;
      if(thisexp.mode == Xmm::IMAGING){
//  	datamodes.push_back("imaging");
//  	evlistcombpars["epndatamodes"]->setValue(datamodes);
	evlistcombpars["imagingset"]->setValue(_proc.imagingeventsname(filtered));
	eventlist = _proc.imagingeventsname(filtered);
      } else if(thisexp.mode == Xmm::TIMING ||
		thisexp.mode == Xmm::BURST){
//  	datamodes.push_back("timing");
//  	evlistcombpars["datamodes"]->setValue(datamodes);
	evlistcombpars["timingset"]->setValue(_proc.timingeventsname(filtered));
	eventlist = _proc.timingeventsname(filtered);
      }
      if(_proc.protectedrun(evlistcomb)){
	Info::Task* dscopyblock = task("dscopyblock");
	ParameterInfo& dscopyblockpars = *dscopyblock->parameters();
	dscopyblockpars.revertToDefault();
	vector<string> blks(1);
	blks[0] = calServer.ccf()->cifName() + ":CALINDEX";
	dscopyblockpars["to"]->setValue(eventlist);
	dscopyblockpars["blocks"]->setValue(blks);
	_proc.protectedrun(dscopyblock);
	if(filtered){
	  Info::Task* evselect = task("evselect");
	  ParameterInfo& evselectpars = *evselect->parameters();
	  evselectpars.revertToDefault();
	  evselectpars["table"]->setValue(eventlist + ":EVENTS");
 	  evselectpars["withfilteredset"]->setValue(false);
	  evselectpars["filteredset"]->setValue(eventlist);
	  evselectpars["destruct"]->setValue(true);
	  evselectpars["expression"]->setValue(_post_selection_expression);
	  evselectpars["writedss"]->setValue(true);
	  evselectpars["updateexposure"]->setValue(true);
	  evselectpars["keepfilteroutput"]->setValue(true);
	  if(!_proc.protectedrun(evselect))
	    Sas::message(Msg::Meta, Msg::Silent, "Post evlistcomb
 filtering on " + eventlist + " failed.");
	}
      } else {
	if(removeeventlists){
	  // Remove the intermediate event lists only if the combination
	  // stage was successful.
	  removeeventlists = false;
	  errstr << "The event list combination stage failed for ";
	  _proc.print(thisexp, errstr);
	  errstr << ". The intermediate event lists will not be removed."
		 << Sas::warning("IntermediateEventListsNotRemoved");
	}
      }
    }
  }
  if(removeeventlists)
    for(unsigned k = 0; k < el.size(); ++k)
      unlink(el[k].c_str());
  
  ENDPROC
}

void Epproc::cleanup() {
  PROC
  if(_mypars["removetemporaries"]->boolValue())
    _proc.remove_temporaries();

  if(_mypars["removeintermediategtis"]->boolValue())
    _proc.remove_gtis();
  ENDPROC
}

Epproc::~Epproc() {
  PROC
  cleanup();
  if(_mypars["removeintermediategtis"]->boolValue())
    _proc.remove_last();
  ENDPROC
}
